#include "env.hpp"
#include "value.hpp"
#include "catch.hpp"
#include <stdexcept>

PTR(Env) Env::empty = NEW(EmptyEnv)();

ExtendedEnv::ExtendedEnv(std::string name, PTR(Val) val, PTR(Env) rest)
{
  this->name = name;
  this->val = val;
  this->rest = rest;
}

bool ExtendedEnv::equals(PTR(Env) other_env)
{
  PTR(ExtendedEnv) other_extended_env = CAST(ExtendedEnv)(other_env);
  if (other_extended_env == nullptr)
    return false;
  else return ((name == other_extended_env->name)
               && val->equals(other_extended_env->val)
               && rest->equals(other_extended_env->rest));
}

PTR(Val) ExtendedEnv::lookup(std::string find_name)
{
#if 1
  if (name == find_name)
    return val;
  else
    return rest->lookup(find_name);
#else
  ExtendedEnv *env = this;
  
  while (1) {
    if (env->name == find_name)
      return env->val;
    else {
      ExtendedEnv* ext_env = dynamic_cast<ExtendedEnv*>(env->rest.get());
      
      if (ext_env == nullptr) {
        throw std::runtime_error("free variable: " + find_name);
      } else {
        env = ext_env;
      }
    }
  }
#endif
}

EmptyEnv::EmptyEnv()
{
}

bool EmptyEnv::equals(PTR(Env) other_env)
{
  return (CAST(EmptyEnv)(other_env) != nullptr);
}

PTR(Val) EmptyEnv::lookup(std::string find_name)
{
  throw std::runtime_error("free variable: " + find_name);
}

TEST_CASE( "Env equals" ) {
  CHECK( (NEW(EmptyEnv)())->equals(Env::empty) );
  
  PTR(Env) x_env = NEW(ExtendedEnv)("x", NEW(NumVal)(10), Env::empty);
  
  CHECK( !Env::empty->equals(x_env) );
  CHECK( x_env->equals(x_env) );
  CHECK( !x_env->equals(Env::empty) );
  CHECK( !x_env->equals(NEW(ExtendedEnv)("x", NEW(NumVal)(11), Env::empty)) );
  CHECK( !x_env->equals(NEW(ExtendedEnv)("y", NEW(NumVal)(10), Env::empty)) );
  CHECK( !x_env->equals(NEW(ExtendedEnv)("x", NEW(NumVal)(10), x_env)) );
}

TEST_CASE( "Lookup" ) {
  PTR(Env) x_env = NEW(ExtendedEnv)("x", NEW(NumVal)(10), Env::empty);
  PTR(Env) y_x_env = NEW(ExtendedEnv)("y", NEW(BoolVal)(true), x_env);

  CHECK_THROWS_WITH( Env::empty->lookup("x"), "free variable: x" );
  CHECK_THROWS_WITH( x_env->lookup("y"), "free variable: y" );

  CHECK( x_env->lookup("x")->equals(NEW(NumVal)(10)) );
  CHECK( y_x_env->lookup("x")->equals(NEW(NumVal)(10)) );
  CHECK( y_x_env->lookup("y")->equals(NEW(BoolVal)(true)) );
}
