#ifndef value_hpp
#define value_hpp

#include "pointer.hpp"

#include <string>

/* A forward declaration, so `Val` can refer to `Expr`, while
   `Expr` still needs to refer to `Val`. */
class Expr;
class Env;

class Val ENABLE_THIS(Val) {
public:
  virtual bool equals(PTR(Val) other_val) = 0;

  // To implement `*` for two values, possibly raising an exception
  virtual PTR(Val) add_to(PTR(Val) other_val) = 0;
  
  // To implement `*` for two values, possibly raising an exception
  virtual PTR(Val) mult_with(PTR(Val) other_val) = 0;

  // To implement `if` test a value, possibly raising an exception
  virtual bool is_true() = 0;
  
  // To implement function call with the value as the called object,
  // possibly raising an exception
  virtual PTR(Val) call(PTR(Val) actual_arg) = 0;

  // To convert a value back to an expression
  // (which is useful for substitution)
  virtual PTR(Expr) to_expr() = 0;
  
  // To print a value to a string
  virtual std::string to_string() = 0;
};

class NumVal : public Val {
public:
  int rep;
  
  NumVal(int rep);
  bool equals(PTR(Val) other_val);

  PTR(Val) add_to(PTR(Val) other_val);
  PTR(Val) mult_with(PTR(Val) other_val);
  bool is_true();
  PTR(Val) call(PTR(Val) actual_arg);

  PTR(Expr) to_expr();
  std::string to_string();
};

class BoolVal : public Val {
public:
  bool rep;
  
  BoolVal(bool rep);
  bool equals(PTR(Val) other_val);

  PTR(Val) add_to(PTR(Val) other_val);
  PTR(Val) mult_with(PTR(Val) other_val);
  bool is_true();
  PTR(Val) call(PTR(Val) actual_arg);

  PTR(Expr) to_expr();
  std::string to_string();
};

class FunVal : public Val {
public:
  std::string formal_arg;
  PTR(Expr) body;
  PTR(Env) env;
  
  FunVal(std::string formal_arg, PTR(Expr) body, PTR(Env) env);
  bool equals(PTR(Val) other_val);

  PTR(Val) add_to(PTR(Val) other_val);
  PTR(Val) mult_with(PTR(Val) other_val);
  bool is_true();
  PTR(Val) call(PTR(Val) actual_arg);

  PTR(Expr) to_expr();
  std::string to_string();
};

#endif /* value_hpp */
